from __future__ import annotations

import datetime
import functools
import logging
import random
from enum import Enum, auto
from typing import TYPE_CHECKING, overload

import discord
from discord import Interaction

from disckit import UtilConfig
from disckit.errors import (
    UnkownCooldownCommand,
    UnkownCooldownContext,
    UnkownCooldownInteraction,
)
from disckit.utils import ErrorEmbed, sku_check_guild, sku_check_user

if TYPE_CHECKING:
    from typing import (
        Any,
        Awaitable,
        Callable,
        Literal,
        Optional,
        ParamSpec,
        Sequence,
        Tuple,
        TypeVar,
        Union,
    )

    P = ParamSpec("P")
    T = TypeVar("T")


__all__ = ("CoolDownBucket", "CoolDown")


logger = logging.getLogger(__name__)


class CoolDownBucket(Enum):
    USER = auto()
    GUILD = auto()
    CHANNEL = auto()


class CoolDown:
    owner_ids: Optional[Sequence[int]] = None
    owner_bypass: bool = False

    cooldown_data: dict[
        CoolDownBucket, dict[str, dict[int, datetime.datetime]]
    ] = {
        CoolDownBucket.USER: {},
        CoolDownBucket.GUILD: {},
        CoolDownBucket.CHANNEL: {},
    }

    @overload
    @staticmethod
    def cooldown(
        time: Union[float, int],
        bucket_type: Literal[CoolDownBucket.USER, CoolDownBucket.GUILD] = ...,
        sku_id: int = ...,
    ) -> Callable[..., Any]: ...

    @overload
    @staticmethod
    def cooldown(
        time: Union[float, int],
        bucket_type: CoolDownBucket = ...,
        sku_id: None = ...,
    ) -> Callable[..., Any]: ...

    @staticmethod
    def cooldown(
        time: Union[float, int],
        bucket_type: CoolDownBucket = CoolDownBucket.USER,
        sku_id: Optional[int] = None,
    ) -> Callable[..., Any]:
        """A command decorator to handle cool downs and cool down replies automatically.

        Parameters
        ----------
        time
            | How long for the cool down to last in seconds.
        bucket_type
            | The bucket type for which the cooldown needs to be in.
        sku_id
            | The SKU ID to check for bypassing the cooldown. Optional and defaults to None.
            | The bucket type needs to be of `CoolDownBucket.USER` if sku_id is supplied.
        """

        def decorator(
            func: Callable[P, Awaitable[Optional[T]]],
        ) -> Callable[P, Awaitable[Optional[T]]]:
            @functools.wraps(func)
            async def wrapper(
                *args: P.args, **kwargs: P.kwargs
            ) -> Optional[T]:
                nonlocal time

                interaction: Optional[Interaction] = None
                cooldown_check: Tuple[bool, Optional[str]] = (True, None)
                sku: bool = False

                for arg in args + tuple(kwargs.values()):
                    if isinstance(arg, discord.Interaction):
                        interaction = arg

                if not isinstance(interaction, discord.Interaction):
                    raise UnkownCooldownInteraction(
                        f"Cannot find the interaction object of the command: {func.__name__}"
                    )

                if CoolDown.owner_bypass:
                    if not CoolDown.owner_ids:
                        logger.error(
                            "`CoolDown.owner_ids` contains falsey data while `CoolDown.owner_bypass` is enabled."
                            "\nHence couldn't determine owner ids to bypass."
                        )
                else:
                    cooldown_check = CoolDown.check(
                        interaction=interaction, bucket_type=bucket_type
                    )

                if sku_id:
                    if bucket_type == CoolDownBucket.USER:
                        sku = await sku_check_user(
                            bot=interaction.client,
                            sku_id=sku_id,
                            user_id=interaction.user.id,
                        )

                    elif bucket_type == CoolDownBucket.GUILD:
                        sku = await sku_check_guild(
                            bot=interaction.client,
                            sku_id=sku_id,
                            guild_id=interaction.user.id,
                        )

                if cooldown_check[0] or sku:
                    CoolDown.add(
                        time=time,
                        interaction=interaction,
                        bucket_type=bucket_type,
                    )
                    return await func(*args, **kwargs)

                cooldown_text = random.choice(
                    UtilConfig.COOLDOWN_TEXTS
                ).format(cooldown_check[1])

                await interaction.response.send_message(
                    embed=ErrorEmbed(cooldown_text), ephemeral=True
                )

            return wrapper

        return decorator

    @staticmethod
    def _get_context(
        bucket_type: CoolDownBucket, interaction: Interaction
    ) -> int:
        primary_context: int

        if bucket_type == CoolDownBucket.USER:
            primary_context = interaction.user.id

        elif bucket_type == CoolDownBucket.GUILD:
            if interaction.guild is None:
                raise UnkownCooldownContext(
                    "Couldn't obtain guild object for bucket type `CoolDownBucket.GUILD`"
                    "\nMake sure the command is being ran in a server."
                )
            primary_context = interaction.guild.id

        elif bucket_type == CoolDownBucket.CHANNEL:
            if interaction.channel is None:
                raise UnkownCooldownContext(
                    "Couldn't obtain channel object for bucket type `CoolDownBucket.CHANNEL`"
                    "\nMake sure the command is being ran in a server."
                )
            primary_context = interaction.channel.id

        return primary_context

    @staticmethod
    def add(
        time: Union[float, int],
        interaction: Interaction,
        bucket_type: CoolDownBucket,
        command_name: Optional[str] = None,
    ) -> None:
        """Adds the cool down to the respective bucket type.

        Parameters
        ----------
        time
            | How long for the cool down to last in seconds.
        interaction
            | The interaction object associated with the command.
        bucket_type
            | The bucket type on which the cooldown should act upon.
        command_name
            | An optional command name to give. This will be choosen over `interaction.command.name` if given.
        """

        if interaction.command is None and command_name is None:
            raise UnkownCooldownCommand("Couldn't determine command name.")

        command = command_name or interaction.command.name  # pyright:ignore[reportOptionalMemberAccess]
        primary_context = CoolDown._get_context(
            bucket_type=bucket_type, interaction=interaction
        )
        current = datetime.datetime.now()

        CoolDown.cooldown_data[bucket_type].setdefault(command, {})
        CoolDown.cooldown_data[bucket_type][command][primary_context] = (
            current + datetime.timedelta(seconds=time)
        )

    @overload
    @staticmethod
    def check(  # pyright:ignore[reportOverlappingOverload]
        interaction: Interaction,
        bucket_type: CoolDownBucket,
        command_name: Optional[str] = None,
        cooldown_return: Literal["string"] = ...,
    ) -> Tuple[bool, Optional[str]]: ...

    @overload
    @staticmethod
    def check(
        interaction: Interaction,
        bucket_type: CoolDownBucket,
        command_name: Optional[str] = None,
        cooldown_return: Literal["datetime"] = ...,
    ) -> Tuple[bool, Optional[datetime.datetime]]: ...

    @staticmethod
    def check(
        interaction: Interaction,
        bucket_type: CoolDownBucket,
        command_name: Optional[str] = None,
        cooldown_return: Literal["datetime", "string"] = "string",
    ) -> Tuple[bool, Optional[Union[str, datetime.datetime]]]:
        """Checks the cooldown for the respective bucket type.

        Parameters
        ----------
        interaction
            | The interaction object associated with the command.
        bucket_type
            | The bucket type on which the cooldown should act upon.
        command_name
            | An optional command name to give. This will be choosen over `interaction.command.name` if given.

        Returns
        -------
        | Returns a tuple where the first element is a bool. If the bool is `False` the user is under cooldown.
        | The second element is either a relative timestamp string or a datetime object of when the cooldown gets
        | over. If the first element `True` meaning no cooldown is imposed, the second element of the tuple will
        | be `None`.

        """

        if interaction.command is None and command_name is None:
            raise UnkownCooldownCommand("Couldn't determine command name.")

        command = command_name or interaction.command.name  # pyright:ignore[reportOptionalMemberAccess]
        primary_context = CoolDown._get_context(
            bucket_type=bucket_type, interaction=interaction
        )
        current = datetime.datetime.now()

        try:
            cooldown = CoolDown.cooldown_data[bucket_type][command][
                primary_context
            ]
        except KeyError:
            return (True, None)

        if current > cooldown:
            try:
                del CoolDown.cooldown_data[bucket_type][command][
                    primary_context
                ]
            except KeyError:
                pass
            finally:
                return (True, None)

        else:
            cooldown = CoolDown.cooldown_data[bucket_type][command][
                primary_context
            ]

            if cooldown_return == "datetime":
                return (False, cooldown)

            cooldown_text = f"<t:{round(cooldown.timestamp())}:R>"
            return (False, cooldown_text)

    @staticmethod
    def reset(
        interaction: Interaction,
        bucket_type: CoolDownBucket,
        command_name: Optional[str] = None,
    ) -> bool:
        """Removes the cool down from the bucket type.

        Parameters
        ----------
        interaction
            | The interaction object associated with the command.
        bucket_type
            | The bucket type on which the cooldown should act upon.
        command_name
            | An optional command name to give. This will be choosen over `interaction.command.name` if given.
        """

        if interaction.command is None and command_name is None:
            raise UnkownCooldownCommand("Couldn't determine command name.")

        command = command_name or interaction.command.name  # pyright:ignore[reportOptionalMemberAccess]
        primary_context = CoolDown._get_context(
            bucket_type=bucket_type, interaction=interaction
        )

        try:
            del CoolDown.cooldown_data[bucket_type][command][primary_context]
            return True
        except KeyError:
            return False



import logging
from typing import Any, Optional

import discord
from discord import Embed, Interaction, app_commands
from discord.ext.commands import Bot
from discord.ui import Select, View

from disckit import UtilConfig
from disckit.cogs import BaseCog
from disckit.utils import MainEmbed, MentionTree
from disckit.utils.paginator import Paginator

logger = logging.getLogger(__name__)


class HelpSelect(Select[Any]):
    def __init__(
        self,
        valid_help_options: list[str],
        cog_embed_data: dict[str, list[Embed]],
    ) -> None:
        self.cog_embed_data = cog_embed_data

        options: list[discord.SelectOption] = [
            discord.SelectOption(
                label=cog.title(),
                value=cog.title(),
            )
            for cog in valid_help_options
        ]

        try:
            del cog_embed_data["Help Cog"]
        except KeyError:
            pass

        super().__init__(
            placeholder="Select A Command Menu",
            options=options,
            min_values=1,
            max_values=1,
        )

    async def callback(self, interaction: Interaction) -> None:
        await interaction.response.defer()

        selected_cog: str = self.values[0]
        if selected_cog == "All Commands":
            all_embeds = []
            for cog, embeds in self.cog_embed_data.items():
                if cog.title() not in UtilConfig.OWNER_ONLY_HELP_COGS:
                    all_embeds.extend(embeds)

        elif selected_cog == "Overview":
            all_embeds = [UtilConfig.OVERVIEW_HELP_EMBED]

        else:
            all_embeds = self.cog_embed_data[selected_cog]

        view = View()
        view.add_item(self)
        paginator = Paginator(
            interaction,
            pages=all_embeds,
            home_page=UtilConfig.OVERVIEW_HELP_EMBED,
            home_view=view,
        )
        message = await interaction.original_response()
        await paginator.start(message)


class HelpCog(BaseCog, name="Help Cog"):
    """The help command based cog."""

    def __init__(self, bot: Bot) -> None:
        super().__init__(logger)
        self.bot: Bot = bot

    def partition_split(self, item: list[Any], chunk_size: int) -> list[Any]:
        return [
            item[i : i + chunk_size] for i in range(0, len(item), chunk_size)
        ]

    async def help_auto_complete(
        self, interaction: Interaction, current: Optional[str]
    ) -> list[app_commands.Choice[str]]:
        cog_copy: list[str] = ["Overview", "All Commands"]
        cog_copy.extend(list(self.bot.cogs.keys()))
        cog_copy = [name.title() for name in cog_copy]

        for cog_name in UtilConfig.IGNORE_HELP_COGS:
            try:
                cog_copy.remove(cog_name.title())
            except ValueError:
                logger.warning(
                    "Couldn't find cog: `%s` under `UtilConfig.IGNORE_HELP_COGS`",
                    cog_name.title(),
                )

        def remove_commands() -> None:
            for cog_name in UtilConfig.OWNER_ONLY_HELP_COGS:
                try:
                    cog_copy.remove(cog_name.title())
                except ValueError:
                    logger.warning(
                        "Couldn't find cog: `%s` under `UtilConfig.OWNER_ONLY_HELP_COGS`",
                        cog_name.title(),
                    )

        if self.bot.owner_id and interaction.user.id != self.bot.owner_id:
            remove_commands()

        elif (
            self.bot.owner_ids
            and interaction.user.id not in self.bot.owner_ids
        ):
            remove_commands()

        commands: list[app_commands.Choice[str]] = [
            app_commands.Choice(name=option.title(), value=option.title())
            for option in cog_copy
        ]
        return commands

    async def get_all_cog_embeds(self) -> dict[str, list[Embed]]:
        tree: MentionTree = self.bot.tree  # pyright:ignore[reportAssignmentType]
        kwargs: dict[str, discord.Object] = {}
        cog_command_description: dict[str, list[str]] = {}
        cog_command_map: dict[str, str] = {}
        cog_command_embed: dict[str, list[Embed]] = {}
        cmd_per_embed: int = 7

        for cog_name, cog_instance in self.bot.cogs.items():
            for command in cog_instance.walk_app_commands():
                if isinstance(command, app_commands.Command):
                    cog_command_map[command.qualified_name] = cog_name

        if UtilConfig.HELP_OWNER_GUILD_ID:
            kwargs["guild"] = discord.Object(UtilConfig.HELP_OWNER_GUILD_ID)

        async for command, mention in tree.walk_mentions(**kwargs):
            cog_command_description.setdefault(
                cog_command_map[command.qualified_name], []
            )
            total_commands = len(
                cog_command_description[
                    cog_command_map[command.qualified_name]
                ]
            )
            cog_command_description[
                cog_command_map[command.qualified_name]
            ].append(
                f"`{total_commands + 1}.` {mention}\n> {command.description}"
            )

        for cog_name, command in cog_command_description.items():
            command_split = self.partition_split(
                cog_command_description[cog_name.title()], cmd_per_embed
            )

            cog_command_embed[cog_name.title()] = []

            for commands in command_split:
                embed = MainEmbed(f"{cog_name.title()}", "\n\n".join(commands))
                cog_command_embed[cog_name.title()].append(embed)

        return cog_command_embed

    @app_commands.command()
    @app_commands.describe(group="The group you want help for.")
    @app_commands.autocomplete(group=help_auto_complete)
    async def help(
        self, interaction: Interaction, group: str = "Overview"
    ) -> None:
        """The bot's help command"""

        await interaction.response.defer()

        group = group.title()
        valid_cogs = await self.help_auto_complete(
            interaction=interaction, current=None
        )
        valid_cog_names: list[str] = [cog.name for cog in valid_cogs]

        required_cog = group if group in valid_cog_names else "Overview"
        requred_embeds = await self.get_all_cog_embeds()

        if required_cog == "All Commands":
            all_embeds = []
            for cog_name, embeds in requred_embeds.items():
                if cog_name.title() not in UtilConfig.OWNER_ONLY_HELP_COGS:
                    all_embeds.extend(embeds)

        elif required_cog == "Overview":
            all_embeds = [UtilConfig.OVERVIEW_HELP_EMBED]

        else:
            all_embeds = requred_embeds[required_cog]

        valid_cog_names.remove("Overview")
        view = View()
        view.add_item(HelpSelect(valid_cog_names, requred_embeds))

        if required_cog == "Overview":
            await interaction.followup.send(
                embed=UtilConfig.OVERVIEW_HELP_EMBED, view=view
            )

        else:
            paginator = Paginator(
                interaction,
                pages=all_embeds,
                home_page=UtilConfig.OVERVIEW_HELP_EMBED,
                home_view=view,
            )
            message = await interaction.original_response()
            await paginator.start(message)


async def setup(bot: Bot) -> None:
    await bot.add_cog(HelpCog(bot))



from __future__ import annotations

from typing import TYPE_CHECKING

from discord.ext import commands
from typing_extensions import override

from disckit.cogs.loader import dis_load_extension

if TYPE_CHECKING:
    from logging import Logger
    from typing import Optional


__all__ = ("dis_load_extension", "BaseCog")


class BaseCog(commands.Cog):
    """The base cog which comes along with basic logging."""

    def __init__(self, logger: Optional[Logger] = None) -> None:
        super().__init__()
        self.logger: Optional[Logger] = logger

    @override
    async def cog_load(self) -> None:
        if self.logger:
            self.logger.info(f"{self.qualified_name} has been loaded.")

    @override
    async def cog_unload(self) -> None:
        if self.logger:
            self.logger.info(f"{self.qualified_name} has been unloaded.")



from __future__ import annotations

import logging
import sys
import traceback
from typing import TYPE_CHECKING

import discord
from discord.ui import Modal, View
from discord.utils import MISSING

from disckit import UtilConfig
from disckit.utils import ErrorEmbed

if TYPE_CHECKING:
    from typing import Any, Optional, Union

    from discord import Interaction, InteractionMessage, Member, Message, User
    from discord.ui import Item


__all__ = ("BaseView", "BaseModal")


logger = logging.getLogger(__name__)


class BaseView(View):
    """A custom base view which extends `discord.ui.View`
    to provide more inbuilt features.

    Parameters
    ----------
    author
        | The author of the `View`. If set to `None` anyone can interact with the `View`.
    timeout
        | In how many seconds the view will timeout.
    disable_on_timeout
        | If set to `True` it will disable all items in the view when it times out.
    stop_on_timeout
        | Stops the view from listening to any further events on timeout.
    """

    def __init__(
        self,
        author: Optional[Union[int, User, Member]] = None,
        timeout: Optional[float] = 180.0,
        disable_on_timeout: bool = True,
        stop_on_timeout: bool = True,
    ) -> None:
        super().__init__(timeout=timeout)

        self.message: Optional[Union[Message, InteractionMessage]] = None

        self._author: Optional[Union[int, User, Member]] = author
        if isinstance(self._author, (discord.User, discord.Member)):
            self._author = self._author.id
        self._disable_on_timeout: bool = disable_on_timeout
        self._stop_on_timeout: bool = stop_on_timeout

    def disable_all_items(self) -> None:
        """Disables all items in the View when called."""

        for item in self.children:
            item.disabled = True  # pyright:ignore[reportAttributeAccessIssue]

    async def on_timeout(self) -> None:
        if self._disable_on_timeout:
            self.disable_all_items()
            if self.message:
                try:
                    await self.message.edit(view=self)
                except discord.errors.NotFound:
                    pass
                except discord.errors.HTTPException as e:
                    if e.code == 50027:
                        logger.error(
                            "Invalid Webhook Token: Unable to edit the message."
                        )
                    elif e.code == 10008:
                        logger.error(
                            "Unknown Message: The message was deleted."
                        )
                    else:
                        raise e
            else:
                raise Warning(
                    f"{traceback.format_exc()}\n\n"
                    f"BaseView.message was not defined in view: {self} to disable the items.",
                )

        if self._stop_on_timeout:
            self.stop()

    async def interaction_check(self, interaction: Interaction) -> bool:
        if self._author is None or interaction.user.id == self._author:
            return True

        await interaction.response.send_message(
            embed=ErrorEmbed("This interaction is not for you!"),
            ephemeral=True,
        )
        return False

    async def on_error(
        self, interaction: Interaction, error: Exception, item: Item[Any]
    ) -> None:
        if not UtilConfig.BUG_REPORT_CHANNEL:
            return await super().on_error(interaction, error, item)

        if interaction.response.is_done():
            await interaction.followup.send(
                embed=ErrorEmbed(
                    "Sorry :(",
                    "An unexpected error has occurred. The developers have been notified of this.",
                )
            )
        await interaction.response.send_message(
            embed=ErrorEmbed(
                "Sorry :(",
                "An unexpected error has occurred. The developers have been notified of this.",
            )
        )

        print(
            f"Ignoring exception in view {item.view or self} for item {item}:",
            file=sys.stderr,
        )
        traceback.print_exception(
            type(error), error, error.__traceback__, file=sys.stderr
        )

        channel = interaction.client.get_channel(
            UtilConfig.BUG_REPORT_CHANNEL
        ) or await interaction.client.fetch_channel(
            UtilConfig.BUG_REPORT_CHANNEL
        )

        frame = (
            error.__traceback__.tb_frame if error.__traceback__ else "Unkown"
        )

        description = (
            "```"
            f"\nError in view-\n{item.view or self}\n"
            f"\nError in item-\n{item}\n"
            f"\nError caused by-\nAuthor Name: {interaction.user}"
            f"\nAuthor ID: {interaction.user.id}\n"
            f"\nError Type-\n{type(error)}\n"
            f"\nError Type Description-\n{frame}\n"
            f"\nCause-\n{error.with_traceback(error.__traceback__)}"
            "```"
        )

        await channel.send(  # pyright:ignore[reportAttributeAccessIssue, reportUnknownMemberType]
            embed=ErrorEmbed(
                "Error caused in a view",
                description,
            )
        )


class BaseModal(Modal):
    """A custom base modal which extends `discord.ui.Modal`
    to provide more inbuilt features.

    Parameters
    -----------
    title
        | The title of the modal.
        | Can only be up to 45 characters.
    timeout
        | Timeout in seconds from last interaction with the UI before no longer accepting input.
        | If ``None`` then there is no timeout.
    custom_id
        | The ID of the modal that gets received during an interaction.
        | If not given then one is generated for you.
        | Can only be up to 100 characters.
    author
        | The author of the modal. Disallows anyone else to use the modal.
    """

    def __init__(
        self,
        *,
        title: str = MISSING,
        timeout: Optional[float] = None,
        custom_id: str = MISSING,
        author: Optional[Union[int, User, Member]] = None,
    ) -> None:
        super().__init__(title=title, timeout=timeout, custom_id=custom_id)

        self._author: Optional[Union[int, User, Member]] = author
        if isinstance(self._author, (discord.User, discord.Member)):
            self._author = self._author.id

    async def interaction_check(self, interaction: Interaction) -> bool:
        if self._author is None or interaction.user.id == self._author:
            return True

        await interaction.response.send_message(
            embed=ErrorEmbed("This interaction is not for you!"),
            ephemeral=True,
        )
        return False

    async def on_error(
        self, interaction: Interaction, error: Exception
    ) -> None:
        if not UtilConfig.BUG_REPORT_CHANNEL:
            await super().on_error(interaction, error)
            return

        if interaction.response.is_done():
            await interaction.followup.send(
                embed=ErrorEmbed(
                    "Sorry :(",
                    "An unexpected error has occurred. The developers have been notified of this.",
                )
            )
        await interaction.response.send_message(
            embed=ErrorEmbed(
                "Sorry :(",
                "An unexpected error has occurred. The developers have been notified of this.",
            )
        )

        print(
            f"Ignoring exception in modal {self}:",
            file=sys.stderr,
        )
        traceback.print_exception(
            type(error), error, error.__traceback__, file=sys.stderr
        )

        channel = interaction.client.get_channel(
            UtilConfig.BUG_REPORT_CHANNEL
        ) or await interaction.client.fetch_channel(
            UtilConfig.BUG_REPORT_CHANNEL
        )

        frame = (
            error.__traceback__.tb_frame if error.__traceback__ else "Unkown"
        )

        description = (
            "```"
            f"\nError in modal-\n{self}\n"
            f"\nError caused by-\nAuthor Name: {interaction.user}"
            f"\nAuthor ID: {interaction.user.id}\n"
            f"\nError Type-\n{type(error)}\n"
            f"\nError Type Description-\n{frame}\n"
            f"\nCause-\n{error.with_traceback(error.__traceback__)}"
            "```"
        )

        await channel.send(  # pyright:ignore[reportAttributeAccessIssue, reportUnknownMemberType]
            embed=ErrorEmbed(
                "Error caused in a modal",
                description,
            )
        )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import discord
from discord.ext import tasks
from typing_extensions import override

from disckit.cogs import BaseCog
from disckit.config import UtilConfig

if TYPE_CHECKING:
    from collections.abc import Iterator

    from discord.ext.commands import Bot

logger = logging.getLogger(__name__)


class StatusHandler(BaseCog, name="Status Handler"):
    """Cog for handling bot's dynamic status."""

    def __init__(self, bot: Bot) -> None:
        super().__init__(logger)
        self.bot: Bot = bot
        self.status: None | Iterator[str] = None

    @override
    async def cog_load(self) -> None:
        self.status_task.start()
        await super().cog_load()

    @override
    async def cog_unload(self) -> None:
        self.status_task.cancel()
        await super().cog_unload()

    @tasks.loop(seconds=UtilConfig.STATUS_COOLDOWN)  # pyright:ignore[reportArgumentType]
    async def status_task(self) -> None:
        await self.bot.wait_until_ready()

        if self.status is None:
            self.status = await self._get_iter()

        try:
            current_status = next(self.status)
        except StopIteration:
            self.status = await self._get_iter()
            current_status = next(self.status)

        await self.bot.change_presence(
            activity=discord.Activity(
                type=UtilConfig.STATUS_TYPE, name=current_status
            )
        )

    async def _get_iter(self) -> Iterator[str]:
        return iter(
            await UtilConfig.STATUS_FUNC[0](
                self.bot, *UtilConfig.STATUS_FUNC[1]
            )
        )


async def setup(bot: Bot) -> None:
    await bot.add_cog(StatusHandler(bot))



from __future__ import annotations

from enum import StrEnum
from typing import TYPE_CHECKING

from discord import ActivityType, ButtonStyle, Colour, Embed

from disckit.utils import default_status_handler

if TYPE_CHECKING:
    from typing import (
        Any,
        ClassVar,
        Coroutine,
        List,
        Optional,
        Protocol,
        Sequence,
        Tuple,
        TypeVar,
        Union,
    )

    from discord import Client

    T_contra = TypeVar("T_contra", bound=Client, contravariant=True)

    class StatusHandlerProtocol(Protocol[T_contra]):
        def __call__(
            self, bot: T_contra, *args: Any
        ) -> Coroutine[Any, Any, Union[Tuple[str, ...], List[str]]]: ...


__all__ = ("UtilConfig", "CogEnum")


_BASE_WORKER_COG_PATH: str = "disckit.cogs.worker."
_BASE_COMMAND_COG_PATH: str = "disckit.cogs.commands."


class UtilConfig:
    """The utility class which configures disckit's utilities.

    Attributes
    ----------
    MAIN_COLOR
        | The color of the MainEmbed.

    SUCCESS_COLOR
        | The color of the SuccessEmbed.

    ERROR_COLOR
        | The color of the ErrorEmbed.

    SUCCESS_EMOJI
        | An emoji used in the title of the SuccessEmbed.

    ERROR_EMOJI
        | An emoji used in the title of the ErrorEmbed.

    FOOTER_IMAGE
        | A URL to an image for the footer of `MainEmbed`, `SuccessEmbed` and `ErrorEmbed`.

    FOOTER_TEXT
        | The footer text of `MainEmbed`, `SuccessEmbed` and `ErrorEmbed`.

    STATUS_FUNC
        | A tuple having its first element as a coroutine object which will be awaited when-
        | - When the cog first loads.
        | - When the handler is done iterating through all statuses returned from the function.
        | The second element is a tuple containing the extra arguments that can be passed to your
        | custom status handler function. If no arguments have to be passed an empty tuple
        | should suffice.

    STATUS_TYPE
        | The discord acitvity type used by the StatusHandler.

    STATUS_COOLDOWN
        | A cooldown in seconds for how long a status will play before changing in the `StatusHandler` cog.

    BUG_REPORT_CHANNEL
        | The channel ID to where the bug reports will be sent to by the `ErrorHandler` cog.

    OWNER_LIST_URL
        | The URL from which to fetch the list of owner IDs for the bot. If not set, no fetching will occur.

    PAGINATOR_BUTTON_STYLE
        | The button style of the paginator buttons.

    PAGINATOR_HOME_BUTTON_STYLE
        | The style for the Home Button in the paginator.

    PAGINATOR_HOME_PAGE_LABEL
        | The label for the home button in the paginator.

    PAGINATOR_HOME_PAGE_EMOJI
        | The emoji for the home button in the paginator.

    PAGINATOR_FIRST_PAGE_EMOJI
        | The emoji for the button controlling the first page in the paginator.

    PAGINATOR_NEXT_PAGE_EMOJI
        | The emoji for the button controlling the next page in the paginator.

    PAGINATOR_PREVIOUS_PAGE_EMOJI
        | The emoji for the button controlling the previous page in the paginator.

    PAGINATOR_LAST_PAGE_EMOJI
        | The emoji for the button controlling the last page in the paginator.

    COOLDOWN_TEXTS
        | The cooldown text used by the cooldown controller.
        | This config needs to have a single placeholder: {}
        | In each of its string elements.

    OWNER_ONLY_HELP_COGS
        | Names of the cogs which are only to be viewed by the owner and
        | not by a regular user in the help command. This only applies to
        | the autocomplete feature while using the command.

    IGNORE_HELP_COGS
        | The names of the cogs to be ignored in the autocomplete feature
        | while running the help command.

    HELP_OWNER_GUILD_ID
        | The guild ID where all commands are synced to for the help command
        | to view them. This includes owner only commands as well.
    """

    def __init__(self) -> None:
        raise RuntimeError("Cannot instantiate UtilConfig.")

    MAIN_COLOR: ClassVar[Optional[Union[int, Colour]]] = 0x5865F2

    SUCCESS_COLOR: ClassVar[Optional[Union[int, Colour]]] = 0x00FF00

    ERROR_COLOR: ClassVar[Optional[Union[int, Colour]]] = 0xFF0000

    SUCCESS_EMOJI: ClassVar[str] = "✅"

    ERROR_EMOJI: ClassVar[str] = "❌"

    FOOTER_IMAGE: ClassVar[Optional[str]] = None

    FOOTER_TEXT: ClassVar[Optional[str]] = None

    STATUS_FUNC: ClassVar[
        Tuple[StatusHandlerProtocol[Any], Tuple[Any, ...]]
    ] = (
        default_status_handler,
        (),
    )

    STATUS_TYPE: ClassVar[ActivityType] = ActivityType.listening

    STATUS_COOLDOWN: ClassVar[Optional[float]] = None

    BUG_REPORT_CHANNEL: ClassVar[Optional[int]] = None

    OWNER_LIST_URL: ClassVar[Optional[str]] = None

    PAGINATOR_BUTTON_STYLE: ClassVar[ButtonStyle] = ButtonStyle.blurple

    PAGINATOR_HOME_BUTTON_STYLE: ClassVar[ButtonStyle] = ButtonStyle.red

    PAGINATOR_HOME_PAGE_LABEL: ClassVar[Optional[str]] = None

    PAGINATOR_HOME_PAGE_EMOJI: ClassVar[Optional[str]] = "🏠"

    PAGINATOR_FIRST_PAGE_EMOJI: ClassVar[str] = "⏪"

    PAGINATOR_NEXT_PAGE_EMOJI: ClassVar[str] = "➡️"

    PAGINATOR_PREVIOUS_PAGE_EMOJI: ClassVar[str] = "⬅️"

    PAGINATOR_LAST_PAGE_EMOJI: ClassVar[str] = "⏩"

    COOLDOWN_TEXTS: Sequence[str] = (
        "Chill, the command will be available {}",
        "What's the hurry? The command will be available {}.",
        "I appreciate your enthusiasm but the command can be used {}.",
        "Take a deep breath in, a deep breath out. The command will be available {}.",
    )

    OWNER_ONLY_HELP_COGS: Sequence[str] = ()

    IGNORE_HELP_COGS: Sequence[str] = (
        "help cog",
        "status handler",
        "error handler",
        "owner id handler",
    )

    OVERVIEW_HELP_EMBED: Embed = Embed(
        title="Bot's Overview",
        description=(
            "Welcome to the help overview of the bot.\n"
            "Use the select menu from below to see the description of different command groups."
        ),
    )

    HELP_OWNER_GUILD_ID: Optional[int] = None


class CogEnum(StrEnum):
    ERROR_HANDLER = _BASE_WORKER_COG_PATH + "error_handler"
    """An extension for error handling."""

    STATUS_HANDLER = _BASE_WORKER_COG_PATH + "status_handler"
    """An extension for the bot's status handling."""

    OWNER_IDS_HANDLER = _BASE_WORKER_COG_PATH + "owner_id_handler"
    """An extension for fetching owner IDs in a URL."""

    HELP_COG = _BASE_COMMAND_COG_PATH + "help"
    """An extension for the help command."""



from __future__ import annotations

import logging
import traceback
from typing import TYPE_CHECKING

import discord
from discord import ButtonStyle, Embed
from discord.ui import Button

from disckit.config import UtilConfig
from disckit.errors import PaginatorInvalidCurrentPage, PaginatorInvalidPages
from disckit.utils import ErrorEmbed
from disckit.utils.ui import BaseModal, BaseView

if TYPE_CHECKING:
    from typing import Any, Optional, Sequence, Union

    from discord import Emoji, Interaction, Message, PartialEmoji
    from discord.ui import TextInput, View


__all__ = ("create_empty_button", "HomeButton", "PageJumpModal", "Paginator")


logger = logging.getLogger(__name__)


def create_empty_button(
    label: str = "\u200b",
    style: ButtonStyle = ButtonStyle.gray,
    disabled: bool = True,
    custom_id: Optional[str] = None,
    emoji: Optional[Union[str, Emoji, PartialEmoji]] = None,
    row: int | None = None,
) -> Button[Any]:
    """Creates a placeholder button with no callback."""

    return Button(
        label=label,
        style=style,
        disabled=disabled,
        custom_id=custom_id,
        emoji=emoji,
        row=row,
    )


class HomeButton(Button["Any"]):
    def __init__(
        self, home_page: Union[str, Embed], new_view: Optional[View] = None
    ) -> None:
        super().__init__(
            emoji=UtilConfig.PAGINATOR_HOME_PAGE_EMOJI,
            label=UtilConfig.PAGINATOR_HOME_PAGE_LABEL,
            style=UtilConfig.PAGINATOR_HOME_BUTTON_STYLE,
        )
        self.home_page: Union[str, Embed] = home_page
        self.new_view: Optional[View] = new_view

    async def callback(self, interaction: Interaction) -> None:
        payload: dict[str, Any] = {"view": self.new_view}
        if isinstance(self.home_page, str):
            payload["content"] = self.home_page
            payload["embed"] = None
        else:
            payload["embed"] = self.home_page
            payload["content"] = None

        await interaction.response.edit_message(**payload)


class PageJumpModal(BaseModal, title="Jump to Page"):
    page_number: TextInput[PageJumpModal] = discord.ui.TextInput(
        label="Enter the page number you want to jump to",
        placeholder="...",
        min_length=1,
        max_length=100,
        style=discord.TextStyle.short,
    )

    def __init__(
        self,
        paginator_view: Paginator,
        author: Optional[int] = None,
    ) -> None:
        super().__init__(author=author)

        self.paginator_view: Paginator = paginator_view
        self.author: Optional[int] = author

        self.page_number.placeholder = f"1 - {self.paginator_view.total_pages}"

    async def on_submit(self, interaction: Interaction) -> None:
        await interaction.response.defer()

        if self.author and interaction.user.id != self.author:
            await interaction.followup.send(
                embed=ErrorEmbed("This interaction is not for you!"),
                ephemeral=True,
            )
            return

        if not self.page_number.value.isdigit():
            await interaction.followup.send(
                embed=ErrorEmbed("Please enter an integer."), ephemeral=True
            )
            return

        page_num = int(self.page_number.value)

        if page_num < 1 or page_num > self.paginator_view.total_pages:
            await interaction.followup.send(
                embed=ErrorEmbed(
                    f"Invalid page number! Please enter a number from 1 - {self.paginator_view.total_pages}"
                ),
                ephemeral=True,
            )
            return

        self.paginator_view.current_page = page_num - 1
        await self.paginator_view.update_paginator(interaction=interaction)


class Paginator(BaseView):
    """A custom, easy-to-use paginator for your bot.
    It can paginate through a list / tuple of strings or embeds.

    Attributes
    ----------
    interaction
        The interaction for the paginator to respond to.
    pages
        The pages to paginate over.
    total_pages
        The total number of pages.
    current_page
        The current page it is on.
    timeout
        The amount of seconds in which the paginator view will time out in.
    author
        The author of the paginator, disallowing anyone else to use it.
    home_page
        Adds a home button if this is supplied
    home_view
        An optional home view which is activated when the home button is used.
    extra_buttons
        Extra buttons to be added to the paginator.
    ephemeral
        A bool for if the paginator needs to be ephemeral or not.
    """

    def __init__(
        self,
        interaction: Interaction,
        *,
        pages: Sequence[Union[str, Embed]],
        start_page: int = 0,
        author: Optional[int] = None,
        home_page: Optional[Union[Embed, str]] = None,
        home_view: Optional[View] = None,
        extra_buttons: Optional[Sequence[Button[Any]]] = None,
        extra_buttons_format: bool = True,
        ephemeral: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Parameters
        ----------
        interaction
            | The interaction for the paginator to respond to.
        pages
            | The pages to paginate over.
        start_page
            | The starting page for the paginator to begin.
        author
            | The author of the paginator, disallowing anyone else to use it.
        home_page
            | Adds a home button if this is supplied
        home_view
            | An optional home view which is activated when the home button is used.
        extra_buttons
            | Extra buttons to be added to the paginator.
        extra_buttons_format
            | If `True` is passed, it formats the buttons in a symmetrical format.
        ephemeral
            | A bool for if the paginator needs to be ephemeral or not.
        """

        super().__init__(author=author, **kwargs)

        self.total_pages: int = len(pages)

        if self.total_pages == 0:
            raise PaginatorInvalidPages(
                "Expected a seqence of 1 or more items (Embed | str). Instead got 0 items."
            )

        if start_page > self.total_pages - 1:
            raise PaginatorInvalidCurrentPage(
                f"Expected an integer of range [0, {len(pages) - 1}]. Instead got {start_page}."
            )

        self.interaction: Interaction = interaction
        self.pages: Sequence[Union[str, Embed]] = pages
        self.current_page: int = start_page
        self.author: Optional[int] = author
        self.home_page: Optional[Union[Embed, str]] = home_page
        self.home_view: Optional[View] = home_view
        self.extra_buttons: list[Button[Any]] = (
            list(extra_buttons) if extra_buttons else []
        )
        self.extra_buttons_format: bool = extra_buttons_format
        self.ephemeral: bool = ephemeral

    def _send_kwargs(self, page_element: Union[Embed, str]) -> dict[str, Any]:
        payload: dict[str, Any] = {"view": self}
        if isinstance(page_element, str):
            payload["content"] = page_element
            payload["embed"] = None
        else:
            payload["embed"] = page_element
            payload["content"] = None
        return payload

    def _add_all_items(self) -> None:
        for button in self.extra_buttons:
            self.add_item(button)

    async def start(self, message: Optional[Message] = None) -> None:
        """Starts the entire paginator

        Parameters
        ----------
        message : Optional[Message]
            If it is not None, the paginator starts by editing this
            message instead of sending a new one
        """

        self.message: Optional[Message] = message

        self.children[
            2
        ].label = f"{self.current_page + 1} / {self.total_pages}"  # pyright:ignore[reportAttributeAccessIssue]

        if self.total_pages == 1:
            for button_index in range(5):
                self.children[button_index].disabled = True  # pyright:ignore[reportAttributeAccessIssue]

        if self.home_page:
            self.add_item(create_empty_button())
            self.add_item(create_empty_button())
            self.add_item(HomeButton(self.home_page, self.home_view))
            self.add_item(create_empty_button())
            self.add_item(create_empty_button())

        total_extra_buttons = len(self.extra_buttons)
        full_rows = 0

        if total_extra_buttons > 0 and self.extra_buttons_format:
            if total_extra_buttons > 5:
                full_rows = total_extra_buttons // 5
                buttons_slice = self.extra_buttons[: full_rows * 5]
                self.extra_buttons = self.extra_buttons[full_rows * 5 :]

                for button in buttons_slice:
                    self.add_item(button)

                total_extra_buttons = len(self.extra_buttons)

            if total_extra_buttons == 1:
                self.add_item(create_empty_button())
                self.add_item(create_empty_button())
                self.add_item(self.extra_buttons[0])
                self.add_item(create_empty_button())
                self.add_item(create_empty_button())

            elif total_extra_buttons % 4 == 0 and total_extra_buttons // 4 < (
                4 - full_rows
            ):
                for button_index in range(0, len(self.extra_buttons) - 1, 4):
                    self.add_item(self.extra_buttons[button_index])
                    self.add_item(self.extra_buttons[button_index + 1])
                    self.add_item(create_empty_button())
                    self.add_item(self.extra_buttons[button_index + 2])
                    self.add_item(self.extra_buttons[button_index + 3])

            elif total_extra_buttons % 3 == 0 and total_extra_buttons // 3 < (
                4 - full_rows
            ):
                for button_index in range(0, len(self.extra_buttons) - 1, 3):
                    self.add_item(create_empty_button())
                    self.add_item(self.extra_buttons[button_index])
                    self.add_item(self.extra_buttons[button_index + 1])
                    self.add_item(self.extra_buttons[button_index + 2])
                    self.add_item(create_empty_button())

            elif total_extra_buttons % 2 == 0 and total_extra_buttons // 2 < (
                4 - full_rows
            ):
                for button_index in range(0, len(self.extra_buttons) - 1, 2):
                    self.add_item(create_empty_button())
                    self.add_item(self.extra_buttons[button_index])
                    self.add_item(create_empty_button())
                    self.add_item(self.extra_buttons[button_index + 1])
                    self.add_item(create_empty_button())

            else:
                self._add_all_items()

        else:
            self._add_all_items()

        element: Union[Embed, str] = self.pages[self.current_page]
        payload_kwargs = self._send_kwargs(element)

        if self.interaction.response.is_done():
            if self.message:
                await self.interaction.followup.edit_message(
                    self.message.id, **payload_kwargs
                )
            else:
                await self.interaction.followup.send(**payload_kwargs)
        else:
            if self.message:
                await self.interaction.response.defer()
                await self.interaction.followup.edit_message(
                    self.message.id, **payload_kwargs
                )

            else:
                await self.interaction.response.send_message(**payload_kwargs)

        if self._disable_on_timeout and not self.message:
            self.message = await self.interaction.original_response()

    async def update_paginator(self, interaction: Interaction) -> None:
        self.children[
            2
        ].label = f"{self.current_page + 1} / {self.total_pages}"  # pyright:ignore[reportAttributeAccessIssue]
        kwargs = self._send_kwargs(self.pages[self.current_page])

        if interaction.response.is_done():
            if not interaction.message:
                await interaction.followup.send(
                    embed=ErrorEmbed("Error!", "Message not found to edit.")
                )
                logger.error("Could not find interaction message to edit.")
                traceback.print_stack()
                return

            await interaction.followup.edit_message(
                interaction.message.id, **kwargs
            )

        else:
            await interaction.response.edit_message(**kwargs)

    @discord.ui.button(
        emoji=UtilConfig.PAGINATOR_FIRST_PAGE_EMOJI,
        style=UtilConfig.PAGINATOR_BUTTON_STYLE,
    )
    async def first_page_callback(
        self, interaction: Interaction, button: Button[Any]
    ) -> None:
        await interaction.response.defer()

        self.current_page = 0

        await self.update_paginator(interaction)

    @discord.ui.button(
        emoji=UtilConfig.PAGINATOR_PREVIOUS_PAGE_EMOJI,
        style=UtilConfig.PAGINATOR_BUTTON_STYLE,
    )
    async def previous_page_callback(
        self, interaction: Interaction, button: Button[Any]
    ) -> None:
        await interaction.response.defer()

        self.current_page -= 1
        if self.current_page < 0:
            self.current_page = self.total_pages - 1
        await self.update_paginator(interaction)

    @discord.ui.button(label="0/0", style=ButtonStyle.gray)
    async def number_page_callback(
        self, interaction: Interaction, button: Button[Any]
    ) -> None:
        await interaction.response.send_modal(PageJumpModal(self, self.author))

    @discord.ui.button(
        emoji=UtilConfig.PAGINATOR_NEXT_PAGE_EMOJI,
        style=UtilConfig.PAGINATOR_BUTTON_STYLE,
    )
    async def next_page_callback(
        self, interaction: Interaction, button: Button[Any]
    ) -> None:
        await interaction.response.defer()

        self.current_page += 1
        if self.current_page > self.total_pages - 1:
            self.current_page = 0
        await self.update_paginator(interaction)

    @discord.ui.button(
        emoji=UtilConfig.PAGINATOR_LAST_PAGE_EMOJI,
        style=UtilConfig.PAGINATOR_BUTTON_STYLE,
    )
    async def last_page_callback(
        self, interaction: Interaction, button: Button[Any]
    ) -> None:
        await interaction.response.defer()

        self.current_page = self.total_pages - 1

        await self.update_paginator(interaction)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from disckit.config import CogEnum, UtilConfig
from disckit.errors import CogLoadError
from disckit.utils import MentionTree

if TYPE_CHECKING:
    from discord.ext.commands import AutoShardedBot, Bot

logger = logging.getLogger(__name__)


async def dis_load_extension(
    bot: AutoShardedBot | Bot,
    *cogs: CogEnum,
    debug_message: bool = True,
) -> None:
    """|coro|
    A custom extension loader specifically for the disckit cogs.

    Parameters
    ----------
    bot
        | The bot instance.

    *cogs
        | The cogs to be loaded from disckit package.

    debug_message
        | The debug message to be printed out when the cog is loaded.
        | Needs to contain one `{}` which is formatted to the cog name
        | being loaded.

    Raises
    ------
    :exc:`CogLoadError`
        | Raised when an error occurrs in loading the cog.
    """

    message = None
    for cog in set(cogs):
        if cog == CogEnum.STATUS_HANDLER:
            if not UtilConfig.STATUS_FUNC:
                message = (
                    "Attribute - `UtilConfig.STATUS_FUNC` needs"
                    " to be set to use StatusHandler cog"
                )

            elif not UtilConfig.STATUS_TYPE:
                message = (
                    "Attribute - `UtilConfig.STATUS_TYPE` needs"
                    "to be set to use StatusHandler cog"
                )

            elif not UtilConfig.STATUS_COOLDOWN:
                message = (
                    "Attribute - `UtilConfig.STATUS_COOLDOWN` needs"
                    " to be set to use StatusHandler cog."
                )

        if cog == CogEnum.ERROR_HANDLER:
            if not UtilConfig.BUG_REPORT_CHANNEL:
                message = (
                    "Attribute - `UtilConfig.BUG_REPORT_CHANNEL` needs"
                    " to be set to use ErrorHandler cog."
                )

        if cog == CogEnum.OWNER_IDS_HANDLER:
            if not UtilConfig.OWNER_LIST_URL:
                message = (
                    "Attribute - `UtilConfig.OWNER_LIST_URL` needs to be"
                    " set to use OwnerIDSHandler cog"
                )

        if cog == CogEnum.HELP_COG:
            if not isinstance(bot.tree, MentionTree):
                message = (
                    "`bot.tree` Needs to be of disckit.utils.MentionTree type. "
                    "Pass in `tree_cls` argument in the initialization as the required type."
                )

        if message:
            raise CogLoadError(message=message, cog=cog)

        await bot.load_extension(cog.value)
        if debug_message:
            logger.info(
                f"Loading extension: {cog.name.title().replace(' ', '')}"
            )



"""
Utility functions for Discord bots – adapted from a gist by LeoCx1000.

The core logic in this file is based on the following public gist:
https://gist.github.com/LeoCx1000/021dc52981299b95ea7790416e4f5ca4

Original Author:
    LeoCx1000 (https://github.com/LeoCx1000)

Modifications and integration into disckit by:
    Jiggly-Balls from the Disutils Team
"""

from __future__ import annotations

from logging import getLogger
from typing import Any, AsyncIterator, Generator, Optional

import discord
from discord import app_commands
from discord.ext import commands

__all__ = ("MentionTree",)
_log = getLogger(__name__)


class MentionTree(app_commands.CommandTree):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.application_commands: dict[
            Optional[int], list[app_commands.AppCommand]
        ] = {}
        self.cache: dict[
            Optional[int],
            dict[
                app_commands.Command[Any, Any, Any]
                | commands.HybridCommand[Any, Any, Any]
                | str,
                str,
            ],
        ] = {}

    async def sync(
        self, *, guild: Optional[discord.abc.Snowflake] = None
    ) -> list[app_commands.AppCommand]:
        """Method overwritten to store the commands."""
        ret = await super().sync(guild=guild)
        guild_id = guild.id if guild else None
        self.application_commands[guild_id] = ret
        self.cache.pop(guild_id, None)
        return ret

    async def fetch_commands(
        self, *, guild: Optional[discord.abc.Snowflake] = None
    ) -> list[app_commands.AppCommand]:
        """Method overwritten to store the commands."""
        ret = await super().fetch_commands(guild=guild)
        guild_id = guild.id if guild else None
        self.application_commands[guild_id] = ret
        self.cache.pop(guild_id, None)
        return ret

    async def get_or_fetch_commands(
        self, *, guild: Optional[discord.abc.Snowflake] = None
    ) -> list[app_commands.AppCommand]:
        """Method overwritten to store the commands."""
        try:
            return self.application_commands[guild.id if guild else None]
        except KeyError:
            return await self.fetch_commands(guild=guild)

    async def find_mention_for(
        self,
        command: app_commands.Command[Any, Any, Any]
        | commands.HybridCommand[Any, Any, Any]
        | str,
        *,
        guild: Optional[discord.abc.Snowflake] = None,
    ) -> Optional[str]:
        """Retrieves the mention of an AppCommand given a specific command name, and optionally, a guild.
        Parameters
        ----------
        name: Union[:class:`app_commands.Command`, :class:`commands.HybridCommand[Any, Any, Any]`, str]
            The command to retrieve the mention for.
        guild: Optional[:class:`discord.abc.Snowflake`]
            The scope (guild) from which to retrieve the commands from. If None is given or not passed,
            only the global scope will be searched, however the global scope will also be searched if
            a guild is passed.

        Returns
        -------
        str | None
            The command mention, if found.
        """

        guild_id = guild.id if guild else None
        try:
            return self.cache[guild_id][command]
        except KeyError:
            pass

        # If a guild is given, and fallback to global is set to True, then we must also
        # check the global scope, as commands for both show in a guild.
        check_global = self.fallback_to_global is True and guild is not None

        if isinstance(command, str):
            # Try and find a command by that name. discord.py does not return children from tree.get_command, but
            # using walk_commands and utils.get is a simple way around that.
            _command = discord.utils.get(
                self.walk_commands(guild=guild), qualified_name=command
            )

            if check_global and not _command:
                _command = discord.utils.get(
                    self.walk_commands(), qualified_name=command
                )

        else:
            _command = command

        if not _command:
            return None

        local_commands = await self.get_or_fetch_commands(guild=guild)
        app_command_found = discord.utils.get(
            local_commands, name=(_command.root_parent or _command).name
        )

        if check_global and not app_command_found:
            global_commands = await self.get_or_fetch_commands(guild=None)
            app_command_found = discord.utils.get(
                global_commands, name=(_command.root_parent or _command).name
            )

        if not app_command_found:
            return None

        mention = f"</{_command.qualified_name}:{app_command_found.id}>"
        self.cache.setdefault(guild_id, {})
        self.cache[guild_id][command] = mention
        return mention

    def _walk_children(
        self,
        commands: list[
            app_commands.Group | app_commands.Command[Any, Any, Any]
        ],
    ) -> Generator[app_commands.Command[Any, Any, Any], None, None]:
        for command in commands:
            if isinstance(command, app_commands.Group):
                yield from self._walk_children(command.commands)
            else:
                yield command

    async def walk_mentions(
        self, *, guild: Optional[discord.abc.Snowflake] = None
    ) -> AsyncIterator[tuple[app_commands.Command[Any, Any, Any], str]]:
        """Gets all valid mentions for app commands in a specific guild.
        This takes into consideration group commands, it will only return mentions for
        the command's children, and not the parent as parents aren't mentionable.

        Parameters
        ----------
        guild: Optional[discord.Guild]
            The guild to get commands for. If not given, it will only return global commands.
        Yields
        ------
        Tuple[Union[:class:`app_commands.Command`, :class:`commands.HybridCommand[Any, Any, Any]`], :class:`str`]

        """
        for command in self._walk_children(
            self.get_commands(
                guild=guild, type=discord.AppCommandType.chat_input
            )
        ):
            mention = await self.find_mention_for(command, guild=guild)
            if mention:
                yield command, mention
        if guild and self.fallback_to_global is True:
            for command in self._walk_children(
                self.get_commands(
                    guild=None, type=discord.AppCommandType.chat_input
                )
            ):
                mention = await self.find_mention_for(command, guild=guild)
                if mention:
                    yield command, mention
                else:
                    _log.warning(
                        "Could not find a mention for command %s in the API. Are you out of sync?",
                        command,
                    )



from __future__ import annotations

from typing import TYPE_CHECKING, overload

from discord import Embed, utils

import disckit

if TYPE_CHECKING:
    from typing import Any, Optional

    import disckit.config


__all__ = ("MainEmbed", "ErrorEmbed", "SuccessEmbed")


class MainEmbed(Embed):
    """Represents a main embed for general use."""

    @overload
    def __init__(self, /, **kwargs: Any) -> None: ...

    @overload
    def __init__(
        self, title: str, /, *, description: str, **kwargs: Any
    ) -> None: ...

    @overload
    def __init__(
        self, description: str, /, *, title: str, **kwargs: Any
    ) -> None: ...

    @overload
    def __init__(self, description: str, /, **kwargs: Any) -> None: ...

    @overload
    def __init__(
        self, title: str, description: str, /, **kwargs: Any
    ) -> None: ...

    def __init__(
        self,
        title: Optional[str] = None,
        description: Optional[str] = None,
        /,
        **kwargs: Any,
    ) -> None:
        """
        Parameters
        ----------
        description
            | The description of the main embed.
        title
            | The title of the main embed.
        """

        if "title" in kwargs and title and not description:
            description = title
        elif not kwargs and not description and title:
            description = title
            title = None
        else:
            description = kwargs.get("description") or description
        title = kwargs.get("title") or title

        kwargs.update({"title": title, "description": description})

        super().__init__(
            color=disckit.config.UtilConfig.MAIN_COLOR,
            timestamp=utils.utcnow(),
            **kwargs,
        )
        self.set_footer(
            text=disckit.config.UtilConfig.FOOTER_TEXT,
            icon_url=disckit.config.UtilConfig.FOOTER_IMAGE,
        )


class ErrorEmbed(Embed):
    """Represents an error embed."""

    @overload
    def __init__(self, /, **kwargs: Any) -> None: ...

    @overload
    def __init__(
        self, title: str, /, *, description: str, **kwargs: Any
    ) -> None: ...

    @overload
    def __init__(
        self, description: str, /, *, title: str, **kwargs: Any
    ) -> None: ...

    @overload
    def __init__(self, description: str, /, **kwargs: Any) -> None: ...

    @overload
    def __init__(
        self, title: str, description: str, /, **kwargs: Any
    ) -> None: ...

    def __init__(
        self,
        title: Optional[str] = None,
        description: Optional[str] = None,
        /,
        **kwargs: Any,
    ) -> None:
        """
        Parameters
        ----------
        description
            | The description of the main embed.
        title
            | The title of the main embed.
        """

        if "title" in kwargs and title and not description:
            description = title
        elif not kwargs and not description and title:
            description = title
            title = None
        else:
            description = kwargs.get("description") or description
        title = kwargs.get("title") or title

        kwargs.update({"title": title, "description": description})

        if kwargs["title"]:
            kwargs["title"] = (
                f"{disckit.config.UtilConfig.ERROR_EMOJI} {kwargs['title']}"
            )

        super().__init__(
            color=disckit.config.UtilConfig.ERROR_COLOR,
            timestamp=utils.utcnow(),
            **kwargs,
        )
        self.set_footer(
            text=disckit.config.UtilConfig.FOOTER_TEXT,
            icon_url=disckit.config.UtilConfig.FOOTER_IMAGE,
        )


class SuccessEmbed(Embed):
    """Represents a success embed."""

    @overload
    def __init__(self, /, **kwargs: Any) -> None: ...

    @overload
    def __init__(
        self, title: str, /, *, description: str, **kwargs: Any
    ) -> None: ...

    @overload
    def __init__(
        self, description: str, /, *, title: str, **kwargs: Any
    ) -> None: ...

    @overload
    def __init__(self, description: str, /, **kwargs: Any) -> None: ...

    @overload
    def __init__(
        self, title: str, description: str, /, **kwargs: Any
    ) -> None: ...

    def __init__(
        self,
        title: Optional[str] = None,
        description: Optional[str] = None,
        /,
        **kwargs: Any,
    ) -> None:
        """
        Parameters
        ----------
        description
            | The description of the main embed.
        title
            | The title of the main embed.
        """

        if "title" in kwargs and title and not description:
            description = title
        elif not kwargs and not description and title:
            description = title
            title = None
        else:
            description = kwargs.get("description") or description
        title = kwargs.get("title") or title

        kwargs.update({"title": title, "description": description})

        if kwargs["title"]:
            kwargs["title"] = (
                f"{disckit.config.UtilConfig.SUCCESS_EMOJI} {kwargs['title']}"
            )

        super().__init__(
            color=disckit.config.UtilConfig.SUCCESS_COLOR,
            timestamp=utils.utcnow(),
            **kwargs,
        )
        self.set_footer(
            text=disckit.config.UtilConfig.FOOTER_TEXT,
            icon_url=disckit.config.UtilConfig.FOOTER_IMAGE,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, final

if TYPE_CHECKING:
    from typing import Any

    from disckit.config import CogEnum


__all__ = (
    "DisException",
    "PaginatorError",
    "CooldownError",
    "CogLoadError",
    "PaginatorInvalidPages",
    "PaginatorInvalidCurrentPage",
    "UnkownCooldownInteraction",
    "UnkownCooldownCommand",
    "UnkownCooldownContext",
)


class DisException(Exception):
    """Base class of disckit's exceptions."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class PaginatorError(DisException):
    """Base exception of all paginator errors."""


class CooldownError(DisException):
    """Base excpetion of all cooldown errors."""


@final
class CogLoadError(DisException):
    """Raised when loading a cog fails.

    Attributes
    ------------
    cog
        The cog that failed loading.
    """

    def __init__(self, message: str, cog: CogEnum, **kwargs: Any) -> None:
        super().__init__(message, **kwargs)
        self.cog = cog


@final
class PaginatorInvalidPages(PaginatorError):
    """Raised when an invalid amount of pages are supplied to the paginator."""


@final
class PaginatorInvalidCurrentPage(PaginatorError):
    """Raised when the current page is invalid of a paginator."""


@final
class UnkownCooldownInteraction(CooldownError):
    """Raises when the cooldown cannot find the interaction object of a command."""


@final
class UnkownCooldownCommand(CooldownError):
    """Raises when the cooldown cannot find the command to act upon."""


@final
class UnkownCooldownContext(CooldownError):
    """Raised when the cooldown doesn't have the required bucket object in the interaction."""



from __future__ import annotations

import ast
import logging
from typing import TYPE_CHECKING

import aiohttp
from discord.ext import tasks
from typing_extensions import override

from disckit.cogs import BaseCog
from disckit.config import UtilConfig

if TYPE_CHECKING:
    from discord.ext.commands import Bot


logger = logging.getLogger(__name__)


class OwnerIDHandler(BaseCog, name="Owner ID Handler"):
    def __init__(self, bot: Bot) -> None:
        super().__init__(logger)
        self.bot: Bot = bot
        self.fetch_owner_ids.start()

    @override
    async def cog_load(self) -> None:
        await super().cog_load()

    @override
    async def cog_unload(self) -> None:
        self.fetch_owner_ids.cancel()
        await super().cog_unload()

    @tasks.loop(hours=12)
    async def fetch_owner_ids(self) -> None:
        url = UtilConfig.OWNER_LIST_URL

        if not url:
            logger.warning("OWNER_LIST_URL is not set in the configuration.")
            return

        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                if response.status == 200:
                    data = await response.text()
                    OWNER_IDS = ast.literal_eval(data)
                    self.bot.owner_ids = set(OWNER_IDS)
                    logger.info("Owner IDs successfully fetched.")
                else:
                    logger.error(
                        f"Failed to fetch owner IDs. Response Status: {response.status}"
                    )

    @fetch_owner_ids.before_loop
    async def before_fetch_owner_ids(self) -> None:
        await self.bot.wait_until_ready()


async def setup(bot: Bot) -> None:
    await bot.add_cog(OwnerIDHandler(bot))



from __future__ import annotations

import functools
from typing import TYPE_CHECKING

import discord
from discord.app_commands import Choice

from disckit.utils.embeds import ErrorEmbed, MainEmbed, SuccessEmbed
from disckit.utils.mention_tree import MentionTree

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable
    from typing import (
        Any,
        Optional,
        ParamSpec,
        Tuple,
        TypeGuard,
        TypeVar,
        Union,
    )

    from discord import Client, Guild, Interaction, Thread, User
    from discord.abc import GuildChannel, PrivateChannel
    from discord.ext.commands import Bot

    T_autocomplete = TypeVar("T_autocomplete", str, int, float)

    P = ParamSpec("P")
    T = TypeVar("T")


__all__ = (
    "MainEmbed",
    "SuccessEmbed",
    "ErrorEmbed",
    "MentionTree",
    "default_status_handler",
    "make_autocomplete",
    "sku_check_guild",
    "sku_check_user",
    "disallow_bots",
    "is_owner",
    "get_or_fetch_guild",
    "get_or_fetch_user",
    "get_or_fetch_channel",
)


async def default_status_handler(bot: Bot, *args: Any) -> Tuple[str, ...]:
    """The default status handler. The first parameter will always be the
    bot instance which will automatically be passed as argument in the
    status handler.

    This function is called when cog first loads and when the handler is
    done iterating through all the statuses returned from the function.


    Parameters
    ----------
    bot
        | The global bot instance that gets passed to the function automatically.
    *args
        | The extra arguments passed in `UtilUtilConfig.STATUS_FUNC[1]`
        | (The second element is the extra arguments that will be passed on).

    Returns
    --------
    | Heehee hawhaw
    """

    users = len(bot.users)
    guilds = len(bot.guilds)
    status = (
        # Prefixed by "Listening to" as the default ActivityType
        # (UtilConfig.STATUS_TYPE = ActivityType.listening).
        f"{users:,} users",
        f"humans from {guilds:,} servers",
        "Slash commands!",
    )

    return status


def type_guarantee(var: Any, *var_types: type[T]) -> TypeGuard[T]:
    """Ensures the type of a variable for type checking purposes only;
    it has no effect at runtime.

    Parameters
    ----------
    var
        The variable to guarentee the type for.
    *var_types
        The types to verify against.
    """

    return True


def make_autocomplete(
    *args: T_autocomplete,
) -> Any:
    """Creates an autocomplete function for the given arguments.
    Returns up to 25 filtered choices based on user input.

    Parameters
    ----------
    *args
        | Options for the autocomplete

    Returns
    --------
    | A function that can be put in @discord.app_commands.autocomplete

    Usage
    ------
    ```
    @app_commands.autocomplete(choice=make_autocomplete("Heads", "Tails"))
    @app_commands.command(name="coin-flip")
    async def coin_flip(
        self, interaction: discord.Interaction, choice: str
    ): ...
    ```
    """
    choices = [Choice(name=str(arg), value=arg) for arg in args]

    async def autocomplete(
        _: Any, current: str
    ) -> list[Choice[T_autocomplete]]:
        if not current:
            return choices[:25]

        return [
            choice
            for choice in choices
            if current.lower() in str(choice.name).lower()
        ][:25]

    return autocomplete


async def sku_check_guild(
    bot: Client,
    sku_id: int,
    *,
    guild_id: int,
) -> bool:
    """|coro|

    Checks if a user or guild has purchased a specific SKU package.
    Only one of user_id or guild_id should be provided.

    Parameters
    ----------
    bot : Client
        The bot class.
    sku_id : int
        The SKU ID of the package.
    guild_id : int
        The Discord guild ID to check

    Returns
    -------
    bool
        True if the user/guild has the entitlement
    """
    sku = discord.Object(id=sku_id)

    guild = discord.Object(id=guild_id)
    guild_entitlements = [
        entitlement
        async for entitlement in bot.entitlements(skus=[sku], guild=guild)
    ]
    return bool(guild_entitlements)


async def sku_check_user(
    bot: Client,
    sku_id: int,
    *,
    user_id: int,
) -> bool:
    """|coro|

    Checks if a user or guild has purchased a specific SKU package.
    Only one of user_id or guild_id should be provided.

    Parameters
    ----------
    bot : Client
        The bot class.
    sku_id : int
        The SKU ID of the package.
    user_id : int
        The Discord user ID to check

    Returns
    -------
    bool
        True if the user/guild has the entitlement
    """
    sku = discord.Object(id=sku_id)

    user = discord.Object(id=user_id)
    user_entitlements = [
        entitlement
        async for entitlement in bot.entitlements(skus=[sku], user=user)
    ]
    return bool(user_entitlements)


def disallow_bots() -> Callable[..., Any]:
    """A decorator used for not allowing members to pass in a bot user into command params"""

    def decorator(
        func: Callable[P, Awaitable[Optional[T]]],
    ) -> Callable[P, Awaitable[Optional[T]]]:
        @functools.wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> Optional[T]:
            interaction: None | Interaction[Bot] = None
            bot_user: bool = False

            for arg in args + tuple(kwargs.values()):
                if isinstance(arg, discord.Interaction):
                    interaction = arg

                elif isinstance(arg, (discord.Member, discord.User)):
                    bot_user = bot_user or arg.bot

            if bot_user and interaction:
                embed = ErrorEmbed("You cannot interact with bots!")

                if not interaction.response.is_done():
                    await interaction.response.send_message(
                        embed=embed, ephemeral=True
                    )
                else:
                    await interaction.followup.send(
                        embed=embed, ephemeral=True
                    )
                return

            return await func(*args, **kwargs)

        return wrapper

    return decorator


def is_owner() -> Callable[..., Any]:
    """A decorator for owner-only slash commands"""

    def decorator(
        func: Callable[P, Awaitable[Optional[T]]],
    ) -> Callable[P, Awaitable[Optional[T]]]:
        """A decorator for owner-only slash commands"""

        @functools.wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> Optional[T]:
            interaction: None | Interaction[Bot] = None

            for arg in args + tuple(kwargs.values()):
                if isinstance(arg, discord.Interaction):
                    interaction = arg
                    break

            if interaction and interaction.client.owner_ids:
                if interaction.user.id in interaction.client.owner_ids:
                    await func(*args, **kwargs)

                else:
                    embed = ErrorEmbed("This command is owner only!")
                    try:
                        await interaction.response.send_message(
                            embed=embed, ephemeral=True
                        )
                    except discord.InteractionResponded:
                        await interaction.followup.send(
                            embed=embed, ephemeral=True
                        )
            else:
                await func(*args, **kwargs)

        return wrapper

    return decorator


async def get_or_fetch_guild(bot: Client, id: int) -> Guild:
    """|coro|

    Tries to fetch the guild object from cache. If it fails, it makes
    an API request to the discord API and return it.

    Parameters
    ----------
    bot
        The bot instance.
    id
        The ID of the guild object you want to obtain.
    """
    return bot.get_guild(id) or await bot.fetch_guild(id)


async def get_or_fetch_user(bot: Client, id: int) -> User:
    """|coro|

    Tries to fetch the user object from cache. If it fails, it makes
    an API request to the discord API and return it.

    Parameters
    ----------
    bot
        The bot instance.
    id
        The ID of the guild object you want to obtain.

    Raises
    ------
    NotFound
        A user with this ID does not exist.
    HTTPException
        Fetching the user failed.
    """

    return bot.get_user(id) or await bot.fetch_user(id)


async def get_or_fetch_channel(
    bot: Client, id: int
) -> Union[PrivateChannel, GuildChannel, Thread]:
    """|coro|

    Tries to fetch the channel object from cache. If it fails, it makes
    an API request to the discord API and return it.

    Parameters
    ----------
    bot
        The bot instance.
    id
        The ID of the channel object you want to obtain.

    Raises
    ------
    InvalidData
        An unknown channel type was received from Discord.
    HTTPException
        Retrieving the channel failed.
    NotFound
        Invalid Channel ID.
    Forbidden
        You do not have permission to fetch this channel.
    """

    return bot.get_channel(id) or await bot.fetch_channel(id)



"""
Disutils' utility package
~~~~~~~~~~~~~~~~~~~~~~~~~

A utility package made for the disutils bots.

:copyright: (c) 2024-present Disutils Team
:license: MIT, see LICENSE for more details.
"""

__version__ = "1.1.1"
__title__ = "disckit"
__author__ = "Jiggly Balls"
__license__ = "MIT"
__copyright__ = "Copyright 2024-present Disutils Team"

from typing import Literal, NamedTuple

from disckit.config import CogEnum, UtilConfig

__all__ = ("UtilConfig", "CogEnum", "version_info")


class _VersionInfo(NamedTuple):
    major: str
    minor: str
    patch: str
    release_level: Literal["alpha", "beta", "final"]


def _expand() -> _VersionInfo:
    v = __version__.split(".")
    level_types = {"a": "alpha", "b": "beta"}
    level = level_types.get(v[-1], "final")
    return _VersionInfo(
        major=v[0],
        minor=v[1],
        patch=v[2],
        release_level=level,  # pyright:ignore[reportArgumentType]
    )


version_info: _VersionInfo = _expand()
